# ******************************************************************************************
# "Dodgem" (A strategy game by Colin Vout, 1972) pygame program by Jos Heinemann 2022
# Play against perfect computer, against another Human or let both sides be played by the
# computer or random" The Perfect play is made possible with analyses from the book
# "Winning Ways for your mathematical plays" by Elwyn R. Berlekamp, John H. Conway and
# Richard K. Guy. Please note, that this programm is mainly meant to show, that it is
# possible to play Dodgem perfectly using the table from Winning Ways. (rules below code)
# ******************************************************************************************
# If you play it perfectly, you are able to beat the computer. Try to find the right moves!

from math import sqrt
from random import choice
from time import time
from sys import exit

import pygame as pg

# init pygame
pg.init()
pg.mixer.init()
pg.font.init()

# init constants
WIN_WIDTH, WIN_HEIGHT = 1920, 1080  # changeable

RED = 255, 0, 0
BLUE = 0, 0, 255
GREEN = 0, 255, 0
WHITE = 255, 255, 255
BLACK = 0, 0, 0

MOVE_WAIT_TIME = 1 # in seconds
next_move_time = time()

# Images
LOGO = pg.image.load("Graphics/logo.png")
DODGEM_TITLE = pg.transform.scale2x(pg.image.load("Graphics/dodgem_logo.png"))
BG = pg.image.load("Graphics/gradient.png")
RED_WON, BLUE_WON = pg.image.load("Graphics/red_won.png"), pg.image.load("Graphics/blue_won.png")

# Sounds
WIN_SOUND = pg.mixer.Sound("Sounds/win.wav")
MOVE_SOUND = pg.mixer.Sound("Sounds/move.wav")
SELECT_SOUND = pg.mixer.Sound("Sounds/select.wav")
LOOSE_SOUND = pg.mixer.Sound("Sounds/loose.wav")

# init pygame screen
display_screen = pg.display.set_mode((WIN_WIDTH*0.5, WIN_HEIGHT*0.5), pg.RESIZABLE)
Screen = pg.Surface((WIN_WIDTH, WIN_HEIGHT))
pg.display.set_caption("Dodgem")
pg.display.set_icon(LOGO)
clock = pg.time.Clock()

Accept_Mouse = True # used to only fire a click event once, when the mouse is held down
Accept_Key = True # used to only fire a key once

class Table:
    # a table where all possible positions are evaluated
    # 0 -> + -> red wins
    # 1 -> - -> blue wins
    # 2 -> * -> next player wins
    # 3 -> o -> next player looses
    table = (
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -
         1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1, 1, -1, -1, 2, 2, -1],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1,
         1, 1, 1, 1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 2, 2, 2],
        [1, -1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 1, 1, -1, 1, 1, -1, 1,
         1, -1, 1, 1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 1, -1, -1, 1, 1, 1, 1, 1, 1, -1, -1, -
         1, 1, 1, 1, 1, 1, -1, 1, 1, -1, -1, -1, -1, 2, 2, -1, 2, -1, -1, 0, 0, -1],
        [1, -1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 1, 1, -1, 1, 1, -1, -1, -
         1, -1, 1, 1, -1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 2, 2, 1, 2, -1, -1, 0, 0, -1],
        [1, -1, -1, 1, 1, -1, 1, 1, -1, -1, -1, -1, 1, -1, -1, 1, 1, -1, 1, 1, -1, 1,
         1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 2, 2, -1, 2, 2, 2, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 1, 1,
         2, 1, 1, 1, 1, 1, 2, -1, -1, -1, 2, 2, 3, 2, 1, 2, -1, 0, -1, 0, -1, 0],
        [1, 1, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1, -1, 1, -1, 1, 1, 1, 1, 1, 1, 1, 1,
         1, 1, 1, 2, 1, -1, 2, 1, -1, 2, 1, -1, 2, 2, -1, 0, 2, 2, 2, 0, 0, 0],
        [-1, 1, -1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 1, 1, -1, 2, 1, -1, 2,
         1, -1, 2, 1, -1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1,
         2, 1, 2, 2, 1, 2, -1, -1, -1, -1, -1, -1, 0, 2, 3, -1, -1, -1, 0, -1, -1],
        [1, 1, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1, -1, 1, -1, 1, 1, 1, 1, 1, 1, -1, -1, -
         1, 2, 1, 1, 2, -1, 1, 2, -1, 1, -1, -1, -1, 0, -1, 2, 0, -1, -1, 0, 0, -1],
        [-1, 1, -1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 1, 1, -1, 1, 1, -1, 1, -1, -
         1, -1, 2, -1, 1, 2, 2, 1, 2, 2, 1, -1, -1, -1, 0, 0, 2, 0, -1, -1, 0, 0, -1],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, 1, -1, -1, 1, 1, 1, -1, -1, -1, 2, 2,
         1, 2, 1, 1, 2, 1, -1, -1, -1, -1, 0, 0, -1, 0, 2, -1, -1, 0, -1, 0, -1, 0],
        [1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1,
         1, 2, 1, 1, 2, -1, -1, 2, -1, -1, 2, -1, -1, 0, -1, -1, 0, 0, 0, 0, 0, 0],
        [-1, 1, -1, 1, -1, 1, 1, -1, 1, -1, -1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1,
         1, -1, 1, 2, -1, 2, 2, 2, -1, 2, 2, -1, 2, 2, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0],
        [1, -1, -1, 1, 1, -1, 1, 1, -1, 1, 1, -1, 1, 2, 2, 1, 1, -1, -1, -1, -1, 2,
         2, -1, 2, 1, -1, 2, 1, 2, -1, -1, -1, 0, 0, 2, 0, 2, 2, -1, 0, -1, 0, -1, 0],
        [1, -1, -1, 1, 1, -1, -1, -1, -1, 1, 1, -1, -1, 2, -1, 1, 1, -1, 1, 1, -1, 1,
         1, -1, 2, 1, -1, 2, -1, 0, 2, -1, 0, 2, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0],
        [-1, -1, -1, 1, -1, -1, 1, -1, -1, 1, -1, -1, 1, 1, 1, 1, -1, -1, 1, -1, -1,
         1, -1, -1, 2, -1, -1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 2, 1, 2, 2, -1, -1, -1, 2, 0, 0, 2, 2,
         0, 2, 1, 2, -1, -1, -1, 2, 0, 0, 0, 0, 2, 3, 2, 0, -1, -1, 0, -1, 0, 0],
        [1, 1, 2, -1, -1, -1, 2, 0, 0, 1, 2, 2, -1, -1, 2, 1, 2, 2, 1, 2, 2, 1, 2,
         2, 1, 2, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0],
        [-1, -1, 0, -1, 1, 2, -1, 0, 2, -1, 0, 2, 1, 1, 1, -1, 0, 0, -1, 0, 0, -1,
         0, 0, -1, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 2, -1, -1, -1, 2, 0, 2, -1, -1, -1,
         1, 2, 1, -1, -1, -1, 0, 0, 0, -1, -1, -1, 2, 0, 2, -1, -1, -1, -1, 0, -1],
        [1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, 1, -1, -1, 1, 1, 2, 1, 1, 2, 1, -1, -1, -
         1, 1, 2, 2, -1, 0, 2, -1, 0, 2, -1, -1, -1, -1, 0, 0, 0, -1, -1, 0, 0, -1],
        [-1, -1, 1, -1, 2, 1, -1, 2, 1, -1, 1, 1, 2, 1, 1, -1, 0, 2, -1, 0, 2, -1, -
         1, -1, -1, 0, 2, 0, 0, 2, 0, 0, 2, -1, -1, -1, 0, 0, 0, 0, -1, -1, 0, 0, -1],
        [1, 1, 1, 1, 1, 1, 1, 2, 1, -1, -1, -1, 2, -1, -1, -1, -1, -1, 2, 0, 2, 2, 1,
         2, 1, 2, 1, -1, -1, -1, 0, 0, -1, 0, 0, -1, 2, 0, -1, -1, -1, 0, -1, 0, 0],
        [1, 1, 1, -1, -1, -1, 1, 1, 1, -1, -1, -1, -1, -1, -1, 1, 2, 1, 1, 2, 1, 1, 2,
         1, 1, 2, 1, -1, 0, -1, -1, 0, -1, -1, 0, -1, -1, 0, -1, 0, 0, 0, 0, 0, 0],
        [-1, -1, 1, -1, 2, 1, -1, 2, 1, -1, -1, -1, 2, -1, -1, -1, 0, 2, -1, 0, 2, -1,
         0, 2, -1, 0, 2, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0],
        [1, -1, -1, 1, 1, -1, 1, 2, -1, 2, 2, -1, 2, 0, 0, -1, -1, -1, 2, 0, -1, 2,
         0, -1, 1, 2, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 2, 0, 0, -1, -1, 0, -1, 0, 0],
        [1, -1, -1, -1, -1, -1, 1, 1, -1, 1, 2, -1, -1, -1, 0, 1, 2, -1, 1, 2, -1, 1,
         2, -1, 1, 2, -1, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0],
        [-1, -1, -1, -1, 2, -1, -1, 2, -1, -1, 0, -1, 2, 2, 2, -1, 0, -1, -1, 0, -1, -
         1, 0, -1, -1, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 1, 1, -1, -1, -1, 2, 2, 1, -1, 2, -1, 2, 1, 1, -1, -1, -1, 0, 0,
         2, 0, 2, 2, 0, -1, 2, -1, -1, -1, 0, -1, 0, 0, -1, 0, -1, 0, -1, 0, -1, 0],
        [-1, 1, -1, 1, -1, 1, 1, -1, 1, 2, -1, 2, 1, 2, 2, 2, -1, 1, -1, -1, -1, 0, -
         1, 0, 0, -1, 2, 0, 2, 2, -1, -1, -1, 0, 0, 0, 0, 0, 0, -1, 0, -1, 0, -1, 0],
        [-1, 1, -1, 1, -1, 2, -1, -1, -1, 2, -1, 3, -1, 0, -1, 2, -1, 2, 2, -1, 2,
         2, -1, 2, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 2, 1, 2, 1, 2, 2, 2, 2, 3, 2, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0,
         2, 2, 2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0],
        [1, 1, 1, -1, -1, -1, 1, 1, 1, 2, 2, 2, -1, -1, 2, 1, 2, 2, -1, -1, -1, 0, 0,
         0, 1, 2, 0, -1, 0, 0, -1, -1, -1, -1, 0, 0, -1, 0, 0, -1, 0, -1, 0, -1, 0],
        [-1, -1, 1, -1, 0, 2, -1, 2, 1, -1, 2, 2, 2, 2, 2, -1, 0, 2, -1, -1, -1, -1,
         0, 0, -1, 0, 0, 0, 0, 0, -1, -1, -1, 0, 0, 0, 0, 0, 0, -1, 0, -1, 0, -1, 0],
        [1, 2, 2, 1, 1, 2, -1, -1, -1, 0, 0, 2, -1, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0,
         0, 2, 1, 2, -1, -1, -1, 0, -1, 0, 0, -1, 0, 0, -1, 0, -1, -1, 0, -1, 0, 0],
        [1, 1, 2, -1, -1, -1, -1, -1, -1, 2, 0, 0, -1, -1, -1, 2, 0, 2, 2, 0, 2, 2, 0,
         2, 1, 1, 2, -1, -1, 0, -1, -1, 0, -1, -1, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0],
        [-1, -1, 2, -1, 0, 2, -1, -1, -1, -1, 0, 2, -1, 0, -1, -1, 0, 0, -1, 0, 0, -1,
         0, 0, -1, 2, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0],
        [-1, 1, -1, 1, -1, 2, 2, -1, 2, 2, -1, 0, 2, 0, 0, -1, -1, -1, 0, -1, 0, 0, -
         1, 0, 2, -1, 0, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, -1, 0, 0],
        [-1, 1, -1, -1, -1, -1, 1, -1, 0, 1, -1, 2, -1, -1, 0, 2, -1, 0, 2, -1, 0,
         2, -1, 0, 0, -1, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0],
        [-1, -1, -1, -1, -1, 0, -1, -1, 0, -1, -1, 0, 0, 0, 0, -1, -1, 0, -1, -1, 0, -
         1, -1, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 2, 2, -1, -1, -1, 2, 2, 2, 0, 0, 0, -1, -1, 0, -1, -1, -1, 0, 0, 0, 0, 0,
         0, 1, 2, 0, -1, -1, -1, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, -1, 0, -1, 0, 0],
        [-1, -1, 2, -1, 0, 0, -1, 0, 2, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, -1,
         0, 0, -1, 0, 0, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, -1, 0, 0],
        [-1, -1, 0, -1, -1, -1, -1, 0, 0, -1, 0, 0, -1, -1, 0, -1, 0, 0, -1, 0, 0, -1,
         0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    top = (
        {(0, 1), (0, 2)},
        {(0, 2), (0, 0)},
        {(0, 1), (0, 0)},
        {(0, 2), (1, 2)},
        {(0, 1), (1, 2)},
        {(1, 2), (0, 0)},
        {(1, 1), (0, 2)},
        {(0, 1), (1, 1)},
        {(1, 1), (0, 0)},
        {(1, 0), (0, 2)},
        {(0, 1), (1, 0)},
        {(1, 0), (0, 0)},
        {(1, 1), (1, 2)},
        {(1, 0), (1, 2)},
        {(1, 0), (1, 1)},
        {(0, 2), (2, 2)},
        {(0, 1), (2, 2)},
        {(2, 2), (0, 0)},
        {(0, 2), (2, 1)},
        {(0, 1), (2, 1)},
        {(2, 1), (0, 0)},
        {(0, 2), (2, 0)},
        {(0, 1), (2, 0)},
        {(2, 0), (0, 0)},
        {(0, 2)},
        {(0, 1)},
        {(0, 0)},
        {(1, 2), (2, 2)},
        {(1, 1), (2, 2)},
        {(1, 0), (2, 2)},
        {(1, 2), (2, 1)},
        {(1, 1), (2, 1)},
        {(1, 0), (2, 1)},
        {(1, 2), (2, 0)},
        {(1, 1), (2, 0)},
        {(1, 0), (2, 0)},
        {(1, 2)},
        {(1, 1)},
        {(1, 0)},
        {(2, 1), (2, 2)},
        {(2, 0), (2, 2)},
        {(2, 0), (2, 1)},
        {(2, 2)},
        {(2, 1)},
        {(2, 0)}
    )
    side = (
        {(2, 0)},
        {(1, 0)},
        {(0, 0)},
        {(1, 0), (2, 0)},
        {(2, 0), (0, 0)},
        {(1, 0), (0, 0)},
        {(2, 1)},
        {(1, 1)},
        {(0, 1)},
        {(2, 0), (2, 1)},
        {(1, 1), (2, 0)},
        {(0, 1), (2, 0)},
        {(1, 0), (2, 1)},
        {(1, 0), (1, 1)},
        {(0, 1), (1, 0)},
        {(2, 1), (0, 0)},
        {(1, 1), (0, 0)},
        {(0, 1), (0, 0)},
        {(2, 2)},
        {(1, 2)},
        {(0, 2)},
        {(2, 0), (2, 2)},
        {(1, 2), (2, 0)},
        {(0, 2), (2, 0)},
        {(1, 0), (2, 2)},
        {(1, 0), (1, 2)},
        {(1, 0), (0, 2)},
        {(2, 2), (0, 0)},
        {(1, 2), (0, 0)},
        {(0, 2), (0, 0)},
        {(1, 1), (2, 1)},
        {(0, 1), (2, 1)},
        {(0, 1), (1, 1)},
        {(2, 1), (2, 2)},
        {(1, 2), (2, 1)},
        {(0, 2), (2, 1)},
        {(1, 1), (2, 2)},
        {(1, 1), (1, 2)},
        {(1, 1), (0, 2)},
        {(0, 1), (2, 2)},
        {(0, 1), (1, 2)},
        {(0, 1), (0, 2)},
        {(1, 2), (2, 2)},
        {(0, 2), (2, 2)},
        {(0, 2), (1, 2)})
    
    @classmethod
    def eval_pos(cls, position:tuple):
        """Returns the eval of the given position, using the position table."""
        # positon --> ({(rx1, ry1), (rx2, ry2)}, {(bx1, by1), (bx2, by2)})
        if len(position[0]) == 0:
            return 0
        elif len(position[1]) == 0:
            return 1
        return cls.table[cls.side.index(position[1])][cls.top.index(position[0])]
    
    @classmethod
    def eval_move(cls, car_idx: int, dx: int = 0, dy: int = 0):
        """Returns the eval of the resulting position after the given move is played."""
        car_positions = list(list(c.position) for c in cars)
        if not (dx or dy): return
        if abs(dx) == 1 and not dy:
            car_positions[car_idx][0] += dx
        elif abs(dy) == 1 and not dx:
            car_positions[car_idx][1] += dy
        else:
            raise ValueError("Cars move one space horizontally or vertically.")
        new_pos = [{tuple(car_positions[i]) for i, c in enumerate(cars) if c.type == Car.RED},
                    {tuple(car_positions[i]) for i, c in enumerate(cars) if c.type == Car.BLUE}]
        # check if cars are outside the field
        if cars[car_idx].position[0] + dx == 3:
            new_pos[0] = {tuple(car_positions[i]) for i, c in enumerate(cars) if c.type == Car.RED and i != car_idx}
        if cars[car_idx].position[1] + dy == -1:
            new_pos[1] = {tuple(car_positions[i]) for i, c in enumerate(cars) if c.type == Car.BLUE and i != car_idx}
        # check if all cars of a side are outside
        if len(new_pos[0]) == 0:
            return 0
        elif len(new_pos[1]) == {}:
            return 1
        return cls.eval_pos(new_pos)

class Car:
    RED = "RED"
    BLUE = "BLUE"
    CAR_SIZE = 125
    class Move_Point:
        MOVE_POINT_SIZE = 90

        @classmethod
        def mouse_over(cls):
            """Decides wther the mouse is over a move point and returns the index."""
            mp = mouse_pos()
            for i, c in enumerate(cars):
                for j, p in enumerate(c.move_points):
                    if d(mp, pixel_position(p.position, 0)) <= p.size:
                        return i, j
            return None
        
        def __init__(self, position, size, car_idx, color=GREEN):
            self.position = position
            self.size = size
            self.car_idx = car_idx
            self.color = color
        
        def draw(self):
            pg.draw.circle(Screen, self.color, pixel_position(self.position, 0), self.size)
        
        def move_car(self):
            c = cars[self.car_idx]
            c.move_points = []
            c.move_to(self.position)
    
    @classmethod
    def check_moves_available(cls):
        # if one side doesn't have a move the side wins
        global Player_move
        player_won = [True, True]
        for c in cars:
            if any(c.find_all_moves()):
                if c.type == Car.BLUE:
                    player_won[0] = False
                else:
                    player_won[1] = False
        if player_won[0] and Player_move:
            win()
            return
        elif player_won[1] and not Player_move:
            win()
            return
        return True
            
    @classmethod
    def mouse_over(cls):
        """Decides wether the mouse is over a car an returns the index."""
        mx, my = mouse_pos()
        for i, c in enumerate(cars):
            pixel_x, pixel_y = pixel_position(c.position, Car.CAR_SIZE)
            if 0 <= mx - pixel_x <= Car.CAR_SIZE and 0 <= my - pixel_y <= Car.CAR_SIZE:
                return i
        return None
    
    def __init__(self, type: str, position: tuple[int]):
        self.type = Car.RED if type.lower() == "red" else Car.BLUE
        self.color = RED if self.type == Car.RED else BLUE
        self.position = position # the position in the field -> (0, 0) or (0, 1) ... or (2, 2)
        self.move_points: list[Car.Move_Point] = []

    def draw(self):
        pixel_pos = pixel_position(self.position, Car.CAR_SIZE)
        pg.draw.rect(Screen, self.color, (pixel_pos[0], pixel_pos[1], Car.CAR_SIZE, Car.CAR_SIZE))
    
    def move_to(self, new_position: tuple):
        global Red_cars, Blue_cars
        """Moves the car to the given position"""
        self.position = new_position
           
        if self.position[0] == Field.size or self.position[1] == -1:
            if self.type == Car.RED:
                Red_cars -= 1
            else:
                Blue_cars -= 1
            cars.remove(self)
            del self

    def move(self, dx: int = 0, dy: int = 0):
        """Moves the car by the given dx and dy values. But only one of them can be 1 or -1 and the other one must be 0."""
        self.move_to((self.position[0] + dx, self.position[1] + dy))

    def find_all_moves(self):
        if self.type == Car.RED:
            # up, right, down, left
            move_directions = [True, True, True, False]
            # check for walls
            if self.position[1] == 0:
                move_directions[0] = False
            elif self.position[1] == Field.size - 1:
                move_directions[2] = False
        else:
            move_directions = [True, True, False, True] # up, right, down, left
            # check for walls
            if self.position[0] == 0:
                move_directions[3] = False
            elif self.position[0] == Field.size - 1:
                move_directions[1] = False
        # check for car standing in the way
        for c in cars:
            if self.position[1] == c.position[1]:
                if c.position[0] == self.position[0] + 1:
                    # can't move right
                    move_directions[1] = False
                elif c.position[0] == self.position[0] - 1:
                    # can't move left
                    move_directions[3] = False
            elif self.position[0] == c.position[0]:
                if c.position[1] == self.position[1] - 1:
                    # can't move up
                    move_directions[0] = False
                elif c.position[1] == self.position[1] + 1:
                    # can't move down
                    move_directions[2] = False
        return tuple(move_directions)

    def create_move_points(self):
        offsets = ((0, -1), (1, 0), (0, 1), (-1, 0))
        for i, dir_possible in enumerate(self.find_all_moves()):
            if dir_possible:
                offset = offsets[i]
                pos = (self.position[0] + offset[0], self.position[1] + offset[1])
                new_point = Car.Move_Point(pos, Car.Move_Point.MOVE_POINT_SIZE, cars.index(self))
                self.move_points.append(new_point)
    
    @classmethod
    def table_move(cls):
        global Player_move
        """Makes the best move found in the tabel."""
        # find all possible moves
        if Player_move:
            curr_type = Car.BLUE
            eval_order = (1, 3, 2, 0, -1)
            # blue wins, next player looses, next player wins, red wins, illegal move
        else:
            curr_type = Car.RED
            eval_order = (0, 3, 2, 1, -1)
            # red wins, next player looses, next player wins, blue wins, illegal move
        move_evals = []
        best_move = 0, 0, 0, eval_order[-1]
        for i, car in enumerate(cars):
            if car.type == curr_type:
                offsets = [(0, -1), (1, 0), (0, 1), (-1, 0)]
                offsets = [o for o, m in zip(offsets, car.find_all_moves()) if m]
                # store eval and move
                eval = [(Table.eval_move(i, dx, dy), (dx, dy)) for dx, dy in offsets]
                move_evals.append((i, eval))
        # find move with best eval
        for i, (car_idx, eval) in enumerate(move_evals):
            for j, e in enumerate(eval):
                if eval_order.index(e[0]) < eval_order.index(best_move[-1]):
                    best_move = car_idx, i, j, e[0] # store new best move
        # make best move
        cars[best_move[0]].move(
            move_evals[best_move[1]][1][best_move[2]][-1][0], move_evals[best_move[1]][1][best_move[2]][-1][1])
        Player_move = not Player_move
    
    @classmethod
    def random_move(cls):
        global Player_move
        """Makes a random possible move."""
        offsets = ((0, -1), (1, 0), (0, 1), (-1, 0))
        curr_type = Car.BLUE if Player_move else Car.RED
        moves_possible = [(i, offsets[j]) for i, c in enumerate(cars) if c.type == curr_type for j, move_is_possible in enumerate(c.find_all_moves()) if move_is_possible]
        if len(moves_possible) == 0: return
        move = choice(moves_possible)
        cars[move[0]].move(move[1][0], move[1][1])
        Player_move = not Player_move

class Field:
    __WIDTH = 750
    size = 3
    cell_size = __WIDTH / size
    __field_positions = [(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1), (0, 2), (1, 2), (2, 2)]
    # helps to draw the lines
    __line_positions = [((1, 0), (1, 3)), ((2, 0), (2, 3)), ((0, 1), (3, 1)), ((0, 2), (3, 2))]
    # centering field
    x = int((WIN_WIDTH - __WIDTH) / 2)
    y = int((WIN_HEIGHT - __WIDTH) / 2)

    __COLOR = WHITE
    __LINE_COLOR = BLACK
    __LINE_WIDTH = 3 # px
    @classmethod
    def set_field(cls, size):
        cls.size = size
        cls.cell_size = cls.__WIDTH / cls.size
        cls.__field_positions = [(x, y) for y in range(size) for x in range(size)]
        cls.__line_positions = [((x, 0), (x, size)) for x in range(1, size)] + [((0, y), (size, y)) for y in range(1, size)]
        Car.CAR_SIZE = 125 / size * 3
        Car.Move_Point.MOVE_POINT_SIZE = 90 / size * 3

    @classmethod
    def draw(cls):
        # (0,0) = top left
        # (2,2) = bottom right
        for x, y in cls.__field_positions: # draw 9 small fields
            pg.draw.rect(Screen, cls.__COLOR, (cls.x + cls.cell_size * x, cls.y + cls.cell_size * y, cls.cell_size, cls.cell_size))
        for (x1, y1), (x2, y2) in cls.__line_positions: # draw lines
            pg.draw.line(Screen, cls.__LINE_COLOR, (cls.x + x1 * cls.cell_size, cls.y + y1 * cls.cell_size),
                         (cls.x + x2 * cls.cell_size, cls.y + y2 * cls.cell_size), cls.__LINE_WIDTH)
    
    @classmethod
    def reset_field(cls):
        global cars, Red_cars, Blue_cars, Player_move
        Red_cars = Blue_cars = cls.size - 1
        Player_move = True
        cars = [Car("red", (0, y)) for y in range(cls.size - 1)] + [Car("blue", (x, cls.size - 1)) for x in range(1, cls.size)]

class Button:
    @classmethod
    def mouse_over(cls):
        """Decides wether the mouse is over a button and returns the index."""
        mx, my = mouse_pos()
        for i, b in enumerate(buttons):
            if 0 <= mx - b.pos[0] <= b.size[0] and 0 <= my - b.pos[1] <= b.size[1]:
                return i
        return None
    
    def __init__(self, name: str, click_function = None, x: float = 0, y: float = 0, w: float = None, h: float = None, positioning = "tl_corner"):
        self.name = name.lower()
        file_path_begin = f"Graphics/Buttons/{self.name}/{self.name}_btn_"
        self.images = [pg.image.load(file_path_begin + path_end + ".png") for path_end in ("std", "hover", "click", "select")]
        self.current_img = self.images[0]
        self.selected = False
        width = self.current_img.get_width() if w is None else w
        height = self.current_img.get_height() if h is None else h
        self.size = width, height
        if positioning == "center":
            # x and y are center position of btn
            x -= width / 2
            y -= height / 2
            self.pos = x, y
        else:
            # x and y are top left corner of btn
            self.pos = x, y
        self.rect = x, y, width, height
        def emty_function(): pass
        self.click_function = emty_function if click_function is None else click_function

    def draw(self):
        Screen.blit(pg.transform.scale(self.current_img, self.size), self.pos)

    def update(self):
        global Accept_Mouse
        idx = buttons.index(self)
        if idx == Button.mouse_over():
            if pg.MOUSEBUTTONUP in event_types:
                self.click_function()
                return True
            if pg.mouse.get_pressed()[0]:
                self.current_img = self.images[2]
                Accept_Mouse = False
            else:
                self.current_img = self.images[1]
        elif self.selected:
            self.current_img = self.images[3]
        else:
            self.current_img = self.images[0]
        return False

    @classmethod
    def select_next(cls):
        for i, b in enumerate(buttons):
            if b.selected:
                b.selected = False
                if i == len(buttons) - 1:
                    buttons[0].selected = True
                else:
                    buttons[i + 1].selected = True
                return
        buttons[0].selected = True
    
    @classmethod
    def press_selected(cls):
        for b in buttons:
            if b.selected:
                b.click_function()
                return
                
    
    @classmethod
    def play_func(cls):
        Game.set_scene("game_mode_select")
    @classmethod
    def simulate_func(cls):
        Game.set_scene("simulate_mode_select")
    @classmethod
    def back_func(cls):
        Game.visited_scenes.pop(-1)
        Game.set_scene(Game.visited_scenes[-1])
        Game.visited_scenes.pop(-1)
    @classmethod
    def computer_func(cls):
        Field.set_field(3)
        Field.reset_field()
        Game.game_mode = 0
        Game.set_scene("game")
    @classmethod
    def human_func(cls):
        Field.set_field(3)
        Field.reset_field()
        Game.game_mode = 1
        Game.set_scene("game")
    @classmethod
    def random_func(cls):
        Field.set_field(3)
        Field.reset_field()
        Game.game_mode = 2
        Game.set_scene("game")
    @classmethod
    def simulate_computer_func(cls):
        global next_move_time
        next_move_time = time() + MOVE_WAIT_TIME
        Field.set_field(3)
        Field.reset_field()
        Game.game_mode = -1
        Game.set_scene("simulate")
    @classmethod
    def simulate_random_func(cls):
        global next_move_time
        next_move_time = time() + MOVE_WAIT_TIME
        Field.set_field(3)
        Field.reset_field()
        Game.game_mode = -2
        Game.set_scene("simulate")
    @classmethod
    def new_game_func(cls):
        global next_move_time
        next_move_time = time() + MOVE_WAIT_TIME
        Field.reset_field()
        Button.back_func()
    @classmethod
    def settings_func(cls):
        Game.set_scene("settings")
    @classmethod
    def main_menu_func(cls):
        Game.set_scene("start")
    @classmethod
    def speed_up_func(cls):
        global MOVE_WAIT_TIME
        MOVE_WAIT_TIME *= 0.75
    @classmethod
    def speed_down_func(cls):
        global MOVE_WAIT_TIME
        MOVE_WAIT_TIME *= 4 / 3
    @classmethod
    def two_by_two_func(cls):
        Field.set_field(8)
        Field.reset_field()
        Game.game_mode = 3
        Game.set_scene("game")

class Game:
    scene = "start"
    visited_scenes = ["start"]
    game_mode = 0

    @classmethod
    def reset(cls):
        cls.scene = "start"
        cls.game_mode = 0

    @classmethod
    def set_scene(cls, value):
        global buttons
        buttons = []
        # place buttons
        match value:
            case "start":
                buttons.append(Button("play", Button.play_func, WIN_WIDTH / 2, 4 * WIN_HEIGHT / 8, None, None, "center"))
                buttons.append(Button("simulate", Button.simulate_func, WIN_WIDTH / 2, 5 * WIN_HEIGHT / 8, None, None, "center"))
                # buttons.append(Button("settings", Button.settings_func, WIN_WIDTH / 2, 6 * WIN_HEIGHT / 8, None, None, "center"))
                buttons.append(Button("quit", exit, WIN_WIDTH / 2, 7 * WIN_HEIGHT / 8, None, None, "center"))
            case "game_mode_select":
                buttons.append(Button("back", Button.back_func, WIN_WIDTH / 10, 9 * WIN_HEIGHT / 10, None, None, "center"))
                buttons.append(Button("computer", Button.computer_func, WIN_WIDTH / 3, WIN_HEIGHT / 3, None, None, "center"))
                buttons.append(Button("human", Button.human_func, 2 * WIN_WIDTH / 3, WIN_HEIGHT / 3, None, None, "center"))
                buttons.append(Button("random", Button.random_func, WIN_WIDTH / 3, 2 * WIN_HEIGHT / 3, None, None, "center"))
                buttons.append(Button("two_by_two", Button.two_by_two_func, 2 * WIN_WIDTH / 3, 2 * WIN_HEIGHT / 3, None, None, "center"))
            case "simulate_mode_select":
                buttons.append(Button("back", Button.back_func, WIN_WIDTH / 10, 9 * WIN_HEIGHT / 10, None, None, "center"))
                buttons.append(Button("computer", Button.simulate_computer_func, WIN_WIDTH / 3, WIN_HEIGHT / 3, None, None, "center"))
                buttons.append(Button("random", Button.simulate_random_func, 2 * WIN_WIDTH / 3, WIN_HEIGHT / 3, None, None, "center"))
            case "game":
                buttons.append(Button("back", Button.back_func, WIN_WIDTH / 10, 9 * WIN_HEIGHT / 10, None, None, "center"))
                buttons.append(Button("reset", Field.reset_field, 9 * WIN_WIDTH / 10, WIN_HEIGHT / 2, None, None, "center"))
            case "simulate":
                buttons.append(Button("back", Button.back_func, WIN_WIDTH / 10, 9 * WIN_HEIGHT / 10, None, None, "center"))
            case "win":
                buttons.append(Button("new_game", Button.new_game_func, WIN_WIDTH / 2, 3 * WIN_HEIGHT / 5, None, None, "center"))
                buttons.append(Button("main_menu", Button.main_menu_func, WIN_WIDTH / 2, 4 * WIN_HEIGHT / 5, None, None, "center"))
            case "settings":
                buttons.append(Button("back", Button.back_func, WIN_WIDTH / 10, 9 * WIN_HEIGHT / 10, None, None, "center"))
                buttons.append(Button("plus", Button.speed_up_func, 3 * WIN_WIDTH / 7, WIN_HEIGHT / 2, None, None, "center"))
                buttons.append(Button("minus", Button.speed_down_func, 4 * WIN_WIDTH / 7, WIN_HEIGHT / 2, None, None, "center"))
        cls.scene = value
        cls.visited_scenes.append(cls.scene)

def mouse_pos():
    """Calculates away the offset of pygame.mouse.get_pos()."""
    lmouse_pos = pg.mouse.get_pos()
    lx = lmouse_pos[0] * (WIN_WIDTH / pg.display.get_window_size()[0])
    ly = lmouse_pos[1] * (WIN_HEIGHT / pg.display.get_window_size()[1])
    return (lx, ly)

def pixel_position(position: tuple[int], obj_size = Car.CAR_SIZE):
    """Returns the pixel coordinates, calculated using the position in the field."""
    return (position[0] * Field.cell_size + (Field.cell_size - obj_size) / 2 + Field.x,
            position[1] * Field.cell_size + (Field.cell_size - obj_size) / 2 + Field.y)


def d(p1, p2):
    return sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)

def check_win():
    if Red_cars == 0 or Blue_cars == 0:
        global Player_move
        Player_move = Blue_cars == 0
        return True
    return False

def win():
    wait_time = 0 # wait time before win screen is shown in seconds (to observe end position)
    stop_until_time = time() + wait_time
    while time() < stop_until_time: pass
    if Game.game_mode == Red_cars == 0:
        pg.mixer.Sound.play(LOOSE_SOUND)
    else:
        pg.mixer.Sound.play(WIN_SOUND)
    Game.set_scene("win")
    return

cars = [Car("red", (0, 0)), Car("red", (0, 1)), Car("blue", (1, 2)), Car("blue", (2, 2))]
buttons: list[Button] = []

Player_move = True  # who's turn (True means blue and False means red)
Red_cars = 2  # num of red cars
Blue_cars = 2  # num of blue cars


# the following lines get the number of possible positions
# empty_fields = sum(row.count(-1) for row in Table.table) # number of impossible positions / empty fields -> (-1)
# all_fields = len(Table.table) * len(Table.table[0])  # get the number of all fields
# print(all_fields - empty_fields) # number of possible positions



def draw_game():
    global Accept_Mouse, Player_move, next_move_time
    match Game.scene:
        case "start" | "win":
            # draw a transparent bg for titel -> more readable
            titel_bg = pg.Surface((DODGEM_TITLE.get_width()+ 10, DODGEM_TITLE.get_height() + 10))
            titel_bg.set_alpha(100)
            titel_bg.fill((255, 255, 255))
            Screen.blit(titel_bg, (WIN_WIDTH / 2 - titel_bg.get_width() / 2, 10))
            # draw DODGEM titel
            Screen.blit(DODGEM_TITLE, (WIN_WIDTH / 2 - DODGEM_TITLE.get_width() / 2, 15))
            # draw buttons
            if Game.scene == "win":
                if Player_move:
                    # BLUE WON!
                    centered_x = WIN_WIDTH / 2 - BLUE_WON.get_width() / 2
                    Screen.blit(BLUE_WON, (centered_x, WIN_HEIGHT / 3))
                else:
                    # RED WON!
                    centered_x = WIN_WIDTH / 2 - RED_WON.get_width() / 2
                    Screen.blit(RED_WON, (centered_x, WIN_HEIGHT / 3))
        case "settings":
            speed_txt_bg = pg.font.SysFont("Calibri", 72, bold=True).render("Speed", False, (0, 0, 0))
            speed_txt = pg.font.SysFont("Calibri", 72, bold=True).render("Speed", False, (255, 255, 255))
            Screen.blit(speed_txt_bg, (WIN_WIDTH / 2 - speed_txt.get_width() / 2 + 3, WIN_HEIGHT / 2 - speed_txt.get_height() / 2))
            Screen.blit(speed_txt_bg, (WIN_WIDTH / 2 - speed_txt.get_width() / 2 - 3, WIN_HEIGHT / 2 - speed_txt.get_height() / 2))
            Screen.blit(speed_txt_bg, (WIN_WIDTH / 2 - speed_txt.get_width() / 2, WIN_HEIGHT / 2 - speed_txt.get_height() / 2 + 3))
            Screen.blit(speed_txt_bg, (WIN_WIDTH / 2 - speed_txt.get_width() / 2, WIN_HEIGHT / 2 - speed_txt.get_height() / 2 - 3))
            Screen.blit(speed_txt, (WIN_WIDTH / 2 - speed_txt.get_width() / 2, WIN_HEIGHT / 2 - speed_txt.get_height() / 2))
        case "game" | "simulate":
            # draw the field
            Field.draw()
            hovered_car = Car.mouse_over()
            hovered_point = Car.Move_Point.mouse_over()
            if check_win():
                win()
                return
            else:
                Car.check_moves_available()

            if Game.scene == "game":
                if not Player_move and Game.game_mode == 0:
                    # computer moves
                    Car.table_move()
                elif not Player_move and Game.game_mode == 2:
                    # random move
                    Car.random_move()
                if pg.MOUSEBUTTONDOWN in event_types and Accept_Mouse:
                    if hovered_car is not None:
                        if (cars[hovered_car].type == Car.BLUE and Player_move) or (cars[hovered_car].type == Car.RED and not Player_move and Game.game_mode in (1, 3)):
                            # car gets selected
                            pg.mixer.Sound.play(SELECT_SOUND)
                            # new move points created
                            for c in cars: c.move_points = []
                            cars[hovered_car].create_move_points()
                            Accept_Mouse = False
                    if hovered_point is not None:
                        # move car to point
                        pg.mixer.Sound.play(MOVE_SOUND)
                        cars[hovered_point[0]].move_points[hovered_point[1]].move_car()
                        Player_move = not Player_move
                        Accept_Mouse = False

            else:
                if time() >= next_move_time:
                    if Game.game_mode == -1:
                        Car.table_move()
                    elif Game.game_mode == -2:
                        Car.random_move()
                    next_move_time = time() + MOVE_WAIT_TIME
            # draw all cars
            for c in cars:
                c.draw()
                for p in c.move_points:
                    p.draw()
           
    for b in buttons:
        if b.update():
            break
        b.draw()

def main():
    global Accept_Mouse, Accept_Key,  event_types
    event_types = []
    run = True
    Game.set_scene("start") 
    while run:
        Screen.fill((20, 20, 20))
        Screen.blit(BG, (0, 0))
        display_screen.fill((0, 0, 0))

        # handle events
        event_types = [event.type for event in pg.event.get()]
        for event in event_types:
            if event == pg.QUIT:
                # stop
                run = False
        # reset the accept mouse if needed
        if not (pg.mouse.get_pressed()[0] or Accept_Mouse): Accept_Mouse = True

        # handle pressed keys
        pressed = pg.key.get_pressed()

        if Accept_Key:
            if pressed[pg.K_TAB]:
                # select next button
                Button.select_next()
                Accept_Key = False

            elif pressed[pg.K_RETURN]:
                # press selected button
                Button.press_selected()
                Accept_Key = False
            
            elif pressed[pg.K_ESCAPE]:
                # exit or go back
                if Game.scene == "start":
                    run = False
                    break
                Button.back_func()
                Accept_Key = False

        elif not any(pressed):
            Accept_Key = True


        # draw frame 
        draw_game()

        display_screen.blit(pg.transform.scale(
            Screen, (display_screen.get_width(), display_screen.get_height())), (0, 0))
        pg.display.update()
        FPS = 30 
        clock.tick(FPS)

if __name__ == "__main__":
    main()
pg.quit()
exit()